
#define len 3
unsigned int nondet_uint();   // returns a non_deterministic uint
_Bool nondet_bool(); 

unsigned int zero_to_n(unsigned int n)
{  
  // constrains the non_detrministic uint to be between 0 and n
  unsigned int result = nondet_uint();
  __CPROVER_assume(result>=0 && result<=n);
  return result;
};


enum state_proc_a_t {a_d, a_i, a_r, a_w, a_cs};
enum state_proc_b_t {b_d, b_i, b_r, b_w, b_cs};
enum state_mo_t {mo_d, mo_i, mo_e};


/* external events for proc_a */
_Bool start_a = 0, next_start_a = 0;
_Bool a_cs_start = 0, next_a_cs_start = 0;
_Bool a_cs_end = 0, next_a_cs_end = 0;

/* external events for proc_b */
_Bool start_b = 0, next_start_b = 0;
_Bool b_cs_start = 0, next_b_cs_start = 0;
_Bool b_cs_end = 0, next_b_cs_end = 0;

/* statechart variables */
_Bool in_cs_a = 0, next_in_cs_a = 0;
_Bool in_cs_b = 0, next_in_cs_b = 0;

_Bool lock = 0, next_lock = 0;
_Bool next_lock_local[3]; //local copy of next lock for each process
_Bool changed[3]; //a  bitt check if proc-i modified shared var
_Bool idle_a = 0, idle_b = 0, idle_m = 0;


/* additional variables to fix the faulty protocol */

_Bool in_wait_a = 0, next_in_wait_a = 0;
_Bool in_wait_b = 0, next_in_wait_b = 0;


unsigned int transition_proc_a (enum state_proc_a_t state_a)
{
  enum state_proc_a_t next_state_a;
  changed[1] = 0; //reset bit to track whether lock is changed
  idle_a = 1;

  switch (state_a)
    {
    case a_d:
      next_state_a = a_i;
      next_in_cs_a = 0;
      idle_a = 0;
      break;

    case a_i:
      if(start_a == 1)
	{
	  next_state_a = a_r;
	  next_start_a = 0;
	  idle_a = 0;
	}
      else
	{
	  next_state_a = a_i;
	}
      break;
      
    case a_r:
      if(a_cs_start == 1)
	{
	  next_state_a = a_w;
	  next_in_wait_a = 1; // added for fixing the protocol
	  next_a_cs_start = 0;
	  idle_a = 0;
	}
      else
	{
	  next_state_a = a_r;
	}
      break;
      
    case a_w:
      // if(lock == 0)
      if((lock == 0) && (in_wait_b == 0)) // modified condition to fix the protocol
	{
	  next_state_a = a_cs;
	  next_in_wait_a = 0; // added for fixing the protocol
	  next_lock_local[1] = 1; changed[1] = 1;
	  next_in_cs_a = 1;
	  idle_a = 0;
	}
      else
	{
	  next_state_a = a_w;
	}
      break;

    case a_cs:
      if(a_cs_end == 1)
	{
	  next_state_a = a_r;
	  next_lock_local[1] = 0; changed[1] = 1;
	  next_in_cs_a = 0;
	  idle_a = 0;
	}
      else
	{
	  next_state_a = a_cs;
	}
      break;

    default :
      // There's an error. One of the states in each chart must be active.
      break;
    }
  
  return next_state_a;
}

unsigned int transition_proc_b (enum state_proc_b_t state_b)
{
  enum state_proc_b_t next_state_b;
  idle_b = 1;
  changed[2] = 0; //reset bit to track whether lock is changed
  
  switch (state_b)
    {
    case b_d:
      next_state_b = b_i;
      next_in_cs_b = 0;
      idle_b = 0;
      break;

    case b_i:
      if(start_b == 1)
	{
	  next_state_b = b_r;
	  next_start_b = 0;
	  idle_b = 0;
	}
      else
	{
	  next_state_b = b_i;
	}
      break;

    case b_r:
      if(b_cs_start == 1)
	{
	  next_state_b = b_w;
	  next_in_wait_b = 1; // added for fixing the protocol
	  next_b_cs_start = 0;
	  idle_b = 0;
	}
      else
	{
	  next_state_b = b_r;
	}
      break;

    case b_w:
      // if(lock == 0)
      if((lock == 0) && (in_wait_a == 0)) // modified condition to fix the protocol
	{
	  next_state_b = b_cs;
	  next_in_wait_b = 0; // added for fixing the protocol
	  next_lock_local[2] = 1; changed[2] = 1;
	  next_in_cs_b = 1;
	  idle_b = 0;
	}
      else
	{
	  next_state_b = b_w;
	}
      break;

    case b_cs:
      if(b_cs_end == 1)
	{
	  next_state_b = b_r;
	  next_lock_local[2] = 0; changed[2] = 1;
	  next_in_cs_b = 0;
	  idle_b = 0;
	}
      else
	{
	  next_state_b = b_cs;
	}
      break;

    default :
      // There's an error. One of the states in each chart must be active.
      break;
    }
  
  return next_state_b;
}

unsigned int transition_mo (enum state_mo_t state_m)
{
  enum state_mo_t next_state_m;
  idle_m = 1;
  changed[0] = 0; //reset bit to track whether lock is changed
  
  switch (state_m)
    {
    case mo_d:
      next_state_m = mo_i;
      next_lock_local[0] = 0; changed[0] = 1;
      idle_m = 0;
      break;
      
    case mo_i:
      if((in_cs_a == 1) &&(in_cs_b == 1))
	{
	  next_state_m = mo_e;
	  idle_m = 0;
	}
      else
	{
	  next_state_m = mo_i;
	}
      break;

    case mo_e:
      next_state_m = mo_e;
      break;

    default :
      // There's an error. One of the states in each chart must be active.
      break;
    }

  return next_state_m;
}

void environment()
{ 
  if((idle_a == 1) && (idle_b == 1) &&(idle_m == 1))
    {
      // assign values to the external events non-deterministically
      start_a = nondet_bool();
      a_cs_start = nondet_bool();
      a_cs_end = nondet_bool();
      
      start_b = nondet_bool();
      b_cs_start = nondet_bool();
      b_cs_end = nondet_bool();

      idle_a = 0; idle_b = 0; idle_m = 0;
    }
  
  return;
}

void update_vars()
{
  // Switch the events off; external as well as internal
  start_a = 0; a_cs_start = 0; a_cs_end = 0;
  start_b = 0; b_cs_start = 0; b_cs_end = 0;
  
  // Copy the values from next_variable to variable
  start_a = next_start_a;
  a_cs_start = next_a_cs_start;
  a_cs_end = next_a_cs_end;

  start_b = next_start_b;
  b_cs_start = next_b_cs_start;
  b_cs_end = next_b_cs_end;

  in_cs_a = next_in_cs_a;
  in_cs_b = next_in_cs_b;

  unsigned int arb_proc;
  arb_proc = zero_to_n(3);
  __CPROVER_assume( ( (!(changed[0]||changed[1]||changed[2]))?(arb_proc==3) :
	                                 ((arb_proc<=2) && changed[arb_proc]) ) );
  if (arb_proc <= 2) lock = next_lock_local[arb_proc];

  /* updating variables to fix the faulty protocol */

  in_wait_a = next_in_wait_a;
  in_wait_b = next_in_wait_b;
  
  return;
}

void main() {
  
  int cnt = 0;
 
  /* active state variables */
  
  enum state_proc_a_t active_state_a; //= a_d;
  enum state_proc_b_t active_state_b; // = b_d;
  enum state_mo_t active_state_m; // = mo_d; */

  _Bool all_idle = ((idle_a == 1) && (idle_b == 1) && (idle_m == 1));
  _Bool acc_all_idle = acc_all_idle || all_idle;

  unsigned int limit = 5;

  while(1)
    {
      
	    if ((cnt >= limit) && !acc_all_idle) assert(0);

	  __CPROVER_assert((cnt>=limit)?acc_all_idle:1, "Stabilizes within limit number of micro-steps");
	  
	    active_state_a = transition_proc_a(active_state_a);
	  
	    active_state_b = transition_proc_b(active_state_b);
	  
	    active_state_m = transition_mo(active_state_m);
	
      update_vars();
      //environment();  do only micro-steps

	  all_idle = ((idle_a == 1) && (idle_b == 1) && (idle_m == 1));
      acc_all_idle = (acc_all_idle || all_idle);
      
      cnt++;
  }
}

