
#define len 3
unsigned int nondet_uint();   // returns a non_deterministic uint
_Bool nondet_bool(); 

enum state_proc_a_t {a_d, a_i, a_r, a_w, a_cs};
enum state_proc_b_t {b_d, b_i, b_r, b_w, b_cs};
enum state_mo_t {mo_d, mo_i, mo_e};

/*
// non-deterministically choose a state.
enum state_proc_a_t nondet_state_proc_a_t();
enum state_proc_b_t nondet_state_proc_b_t();
enum state_mo_t nondet_state_mo_t();
*/

/*
// though the notion of a composite state isn't needed, it may be defined this way
typedef struct composite_state_s {
  enum state_proc_a_t state_a;
  enum state_proc_b_t state_b;
  enum state_mo_t state_m;
} composite_state_t;
*/

/* external events for proc_a */
_Bool start_a = 0, next_start_a = 0;
_Bool a_cs_start = 0, next_a_cs_start = 0;
_Bool a_cs_end = 0, next_a_cs_end = 0;

/* external events for proc_b */
_Bool start_b = 0, next_start_b = 0;
_Bool b_cs_start = 0, next_b_cs_start = 0;
_Bool b_cs_end = 0, next_b_cs_end = 0;

/* statechart variables */
_Bool in_cs_a = 0, next_in_cs_a = 0;
_Bool in_cs_b = 0, next_in_cs_b = 0;

_Bool lock = 0, next_lock = 0;
_Bool idle_a = 0, idle_b = 0, idle_m = 0;

unsigned int transition_proc_a (enum state_proc_a_t state_a)
{
  enum state_proc_a_t next_state_a;
  idle_a = 1;

  switch (state_a)
    {
    case a_d:
      next_state_a = a_i;
      next_in_cs_a = 0;
      idle_a = 0;
      break;

    case a_i:
      if(start_a == 1)
	{
	  next_state_a = a_r;
	  next_start_a = 0;
	  idle_a = 0;
	}
      else
	{
	  next_state_a = a_i;
	}
      break;

    case a_r:
      if(a_cs_start == 1)
	{
	  next_state_a = a_w;
	  next_a_cs_start = 0;
	  idle_a = 0;
	}
      else
	{
	  next_state_a = a_r;
	}
      break;
      
    case a_w:
      if(lock == 0)
	{
	  next_state_a = a_cs;
	  next_lock = 1;
	  next_in_cs_a = 1;
	  idle_a = 0;
	}
      else
	{
	  next_state_a = a_w;
	}
      break;

    case a_cs:
      if(a_cs_end == 1)
	{
	  next_state_a = a_r;
	  next_lock = 0;
	  next_in_cs_a = 0;
	  idle_a = 0;
	}
      else
	{
	  next_state_a = a_cs;
	}
      break;
      
    default :
      // There's an error. One of the states in each chart must be active.
      break;
    }
  
  return next_state_a;
}

unsigned int transition_proc_b (enum state_proc_b_t state_b)
{
  enum state_proc_b_t next_state_b;
  idle_b = 1;

  switch (state_b)
    {
    case b_d:
      next_state_b = b_i;
      next_in_cs_b = 0;
      idle_b = 0;
      break;
      
    case b_i:
      if(start_b == 1)
	{
	  next_state_b = b_r;
	  next_start_b = 0;
	  idle_b = 0;
	}
      else
	{
	  next_state_b = b_i;
	}
      break;
      
    case b_r:
      if(b_cs_start == 1)
	{
	  next_state_b = b_w;
	  next_b_cs_start = 0;
	  idle_b = 0;
	}
      else
	{
	  next_state_b = b_r;
	}
      break;

    case b_w:
      if(lock == 0)
	{
	  next_state_b = b_cs;
	  next_lock = 1;
	  next_in_cs_b = 1;
	  idle_b = 0;
	}
      else
	{
	  next_state_b = b_w;
	}
      break;
      
    case b_cs:
      if(b_cs_end == 1)
	{
	  next_state_b = b_r;
	  next_lock = 0;
	  next_in_cs_b = 0;
	  idle_b = 0;
	}
      else
	{
	  next_state_b = b_cs;
	}
      break;

    default :
      // There's an error. One of the states in each chart must be active.
      break;
    }
  
  return next_state_b;
}

unsigned int transition_mo (enum state_mo_t state_m)
{
  enum state_mo_t next_state_m;
  idle_m = 1;
  
  switch (state_m)
    {
    case mo_d:
      next_state_m = mo_i;
      next_lock = 0;
      idle_m = 0;
      break;
      
    case mo_i:
      if((in_cs_a == 1) && (in_cs_b == 1))
	{
	  next_state_m = mo_e;
	  idle_m = 0;
	}
      else
	{
	  next_state_m = mo_i;
	}
      break;
      
    case mo_e:
      next_state_m = mo_e;
      break;
      
    default :
      // There's an error. One of the states in each chart must be active.
      break;
    }
  
  return next_state_m;
}

void environment()
{
  // This would not work fine when delays are involved.

  /*
    The delay abstraction, as in SAL, can be implemented though.
    Delays are treated as boolean variables, set as the beginning of a superstep
    in the same manner as external events.

    The clock is advanced by a finite amount and all delays which qualify are set as true.

    This is incomplete. There are behaviors allowed by Statemate semantics which are not captured
    through this implementation.
   */
  
  if((idle_a == 1) && (idle_b == 1) &&(idle_m == 1))
    {
      // assign values to the external events non-deterministically
      start_a = nondet_bool();
      a_cs_start = nondet_bool();
      a_cs_end = nondet_bool();
      
      start_b = nondet_bool();
      b_cs_start = nondet_bool();
      b_cs_end = nondet_bool();

      idle_a = 0; idle_b = 0; idle_m = 0;
    }
  
  return;
}

void update_vars()
{
  // Switch the events off; external as well as internal
  start_a = 0; a_cs_start = 0; a_cs_end = 0;
  start_b = 0; b_cs_start = 0; b_cs_end = 0;
  
  // Copy the values from next_variable to variable
  start_a = next_start_a;
  a_cs_start = next_a_cs_start;
  a_cs_end = next_a_cs_end;

  start_b = next_start_b;
  b_cs_start = next_b_cs_start;
  b_cs_end = next_b_cs_end;

  in_cs_a = next_in_cs_a;
  in_cs_b = next_in_cs_b;

  lock = next_lock;

  return;
}

unsigned int zero_to_n(unsigned int n)
{  
  // constrains the non_detrministic uint to be between 0 and n
  unsigned int result = nondet_uint();
  __CPROVER_assume(result>=0 && result<=n);
  return result;
};

/*
unsigned int zero_to_n_not_m(unsigned int n, unsigned int m)
{ 
  // returns a non-determnistic uint <=n that is distinct from m
  unsigned int result=nondet_uint();
  __CPROVER_assume(result>=0 && result<=n && result!=m);
  return result;
};
*/

void main() {
  int i,j,k,l;
  
  int cnt = 0;
  int proc_index_order[len];
  
  k = zero_to_n(len-1);
  l = zero_to_n(len-1);

  /* active state variables */
  enum state_proc_a_t active_state_a = a_d;
  enum state_proc_b_t active_state_b = b_d;
  enum state_mo_t active_state_m = mo_d;
  
  while(1)
    {
      // non-deterministically generate an arbitrary process index permutation
      for(i=0; i<len; i++)
	{
	  proc_index_order[i] = zero_to_n(len-1);
	  for(j=0; j<i; j++)
	    __CPROVER_assume(proc_index_order[j] != proc_index_order[i]);
	}
      
      /*
      // Check assertions about generated permutations
      if ((k!=l) && (cnt == 2))  //choose cnt value to however many times you want to  iterate over while loop
	__CPROVER_assert(proc_index_order[k] == proc_index_order[l], "Expected to fail");
      */
      
      if (k!=l)
	__CPROVER_assert(proc_index_order[k] != proc_index_order[l] &&
			 proc_index_order[k]<3 && proc_index_order[l]<3, 
			 "Expected to pass");
      
      /*  Include your code to execute statechart processes in the generated order
	  for one step inside a superstep*/
      
      for(i=0; i<len; i++)
	{
	  if(proc_index_order[i] == 0)
	    active_state_a = transition_proc_a(active_state_a);
	  if(proc_index_order[i] == 1)
	    active_state_b = transition_proc_b(active_state_b);
	  if(proc_index_order[i] == 2)
	    active_state_m = transition_mo(active_state_m);
	}
      update_vars();
      environment();

      // asserting unreachability of the error state in mo
      __CPROVER_assert(active_state_m != mo_e, "error state reached\n");
      
      cnt++;
    }
}

/*
  Use following CBMC command;  the unwind number must be greater (8-10) than what is required to
  unwind the for looks completely to generate a permutation completely
  I have redireted the transcript to an output file so you can review it later
  Review assignments to array elements in each iteratopn of the while loop to confirm result

  hw-cbmc.exe <file-name> --main.1 (or main.2) --unwind number --no-unwinding-assertions >transcript.txt
  
*/
